use rand::Rng;
use sdl2::{event::Event, keyboard::Keycode, render::Canvas, video::Window};
use std::{time::Duration, time::Instant};

mod doodad;
mod inputstatus;
mod physics;
mod ship;
mod spriteatlas;

use crate::{
	doodad::Doodad, inputstatus::InputStatus, physics::PhysVec, physics::Position, ship::Ship,
	spriteatlas::SpriteAtlas,
};

trait Drawable {
	/// Return the world position of this object
	fn position(&self) -> PhysVec;

	/// Get the position of this drawable on the screen
	/// (0, 0) is at top-left corner.
	///
	/// Returned position is this object's center.
	fn screen_position(&self, canvas: &mut Canvas<Window>, c: &Camera) -> PhysVec {
		let win_size = PhysVec::from(canvas.window().size());
		let h = win_size / 2.0;
		return self.position() - c.pos + h;
	}

	// Draw this item on the screen
	fn draw(&self, canvas: &mut Canvas<Window>, sa: &SpriteAtlas, c: &Camera)
		-> Result<(), String>;
}

enum ShipKind {
	Gypsum,
}

impl ShipKind {
	fn sprite(&self) -> &'static str {
		match self {
			Self::Gypsum => "gypsum.png",
		}
	}
}

struct Camera {
	pos: PhysVec,
}

impl Camera {
	fn new() -> Self {
		Camera {
			pos: PhysVec { x: 0.0, y: 0.0 },
		}
	}
}

static FTL: f64 = 1.0 / 200.0; // frame time limit

struct System {
	bodies: Vec<Box<dyn Drawable>>,
}

impl System {
	fn new() -> Self {
		let mut s = System { bodies: Vec::new() };

		let mut num = rand::thread_rng();
		for _ in 0..25 {
			s.bodies.push(Box::new(Doodad {
				sprite: "small.png".to_owned(),
				pos: Position::new_cartesian(
					num.gen_range(-500.0..500.0),
					num.gen_range(-500.0..500.0),
				),
				scale: 1,
				angle: num.gen_range(-3.0..3.0),
			}));
		}

		s.bodies.push(Box::new(Doodad {
			pos: Position::new_polar(PhysVec { x: 0.0, y: 0.0 }, 0.0, 0.0),
			sprite: "a0.png".to_owned(),
			scale: 1,
			angle: 0.0,
		}));

		s.bodies.push(Box::new(Doodad {
			pos: Position::new_polar(PhysVec { x: 0.0, y: 0.0 }, 300.0, 31.0),
			sprite: "earth.png".to_owned(),
			scale: 1,
			angle: 180.0,
		}));

		return s;
	}

	/// Calculate the state of this body after t seconds.
	pub fn tick(&mut self, _t: f64) {
		//let body = &mut self.bodies[1];
		//body.pos.angle += 0.1 * t;
		//body.angle -= 0.1 * t;
	}

	fn draw(
		&self,
		canvas: &mut Canvas<Window>,
		sa: &SpriteAtlas,
		c: &Camera,
	) -> Result<(), String> {
		for body in &self.bodies {
			body.draw(canvas, sa, c)?;
		}
		return Ok(());
	}
}

fn main() -> Result<(), String> {
	let sdl_context = sdl2::init()?;
	let video_subsystem = sdl_context.video()?;

	let window = video_subsystem
		.window("SDL2", 640, 480)
		.position_centered()
		.resizable()
		.build()
		.map_err(|e| e.to_string())?;

	let mut canvas = window
		.into_canvas()
		.accelerated()
		.build()
		.map_err(|e| e.to_string())?;
	let texture_creator = canvas.texture_creator();

	let sa = SpriteAtlas::new(&texture_creator)?;
	let mut event_pump = sdl_context.event_pump()?;

	let mut system = System::new();

	let mut i = InputStatus::new();
	let mut c = Camera::new();

	let mut s = Ship::new(ShipKind::Gypsum);

	//let mut last_frame_time = 0f64;
	let mut frame_start;

	let mut running = true;
	while running {
		frame_start = Instant::now();

		for event in event_pump.poll_iter() {
			match event {
				Event::Quit { .. }
				| Event::KeyDown {
					keycode: Some(Keycode::Escape),
					..
				} => {
					running = false;
				}
				_ => i.update(event),
			}
		}

		c.pos = s.body.pos;

		// Draw
		canvas.clear();
		system.draw(&mut canvas, &sa, &c)?;
		s.draw(&mut canvas, &sa, &c)?;
		canvas.present();

		let frame_time = frame_start.elapsed().as_secs_f64();

		// Wait
		// (limit frame rate)
		if frame_time < FTL {
			std::thread::sleep(Duration::from_secs_f64(FTL - frame_time));
		}

		// Apply input

		let frame_time = frame_start.elapsed().as_secs_f64();
		//last_frame_time = frame_time;

		s.body.tick(frame_time);
		system.tick(frame_time);

		if i.key_thrust {
			s.body.thrust(50.0 * frame_time);
		}

		if i.key_right {
			s.body.rot(-1.0 * frame_time);
		}

		if i.key_left {
			s.body.rot(1.0 * frame_time);
		}
	}

	Ok(())
}
